/*
	Copyright (c) 2010 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/

#ifndef SYMVEC_H
#define SYMVEC_H

#include <string>
#include <sstream>
#include "symmath.h"

/*
TODO - atm vecs only allow one type
however expressions are typed by their tree
this means vector functions cannot have functions of different expression trees
(i.e. no v = vec(x, y+z, x*z); )
this could be circumvented by making a wrapper class that wrapped any expression, but that would make recasting the expression difficult ...
it might be safest to just create a new class ... 

TODO
r = (x,y,z) <=> r.x = x, r.y = y, r.z = z
v = (f(r), g(r), h(r)) <=> v.x = f(r), v.y = g(r), v.z = h(r)
vector operators
div(v,r) = diff(v.x, r.x) + diff(v.y, r.y) + diff(v.z, r.z)
curl(v,r) = vec( diff(v.z, r.y) - diff(v.y, r.z), ... )		
... if we implemented diff() as a class with a multiply operator (call it 'partial') then we might be able to shortcut these...
... div(v,r) = dot(partial(r), v)
... curl(v,r) = cross( vec(partial(r.x), partial(r.y), partial(r.z)), v) = cross( partial(r), v)
...	... that would mean ensuring partial(x) * x*x*x evaluates right-to-left (so the partial() handles the whole expr)
		... which it doesn't ...
		
... and matrix support: diff(v,r) = J
... ... and tensor support? diff(J,r) = T
... ... ... and rewriting the matrix library operations to just be vectors of vectors, to further tensor support?
*/

namespace symmath {

template<int N, typename T, typename U>
class vec_op;

template<int I, int N, typename T, typename U>
class vec_nth_elem_type;

template<int I, int N, typename T, typename U>
inline typename vec_nth_elem_type<I,N,T,U>::type &vec_elem(vec_op<N,T,U> &v);

//so vec v;
//v.next.next.....last (next N-1 times) is the first elem
//v.last is the N'th elem
//...and on and on

template<int N, typename T, typename U>
class vec_op {
public:
	enum { dim = N };
	typedef T nextType;
	typedef U lastType;
	T next;
	U last;
	
	typedef vec_op<N, T, U> stype;
	typedef vec_op<N, typename T::dtype, typename U::dtype> dtype;
	
	//constructor
	vec_op(const T &next_, const U &last_) : next(next_), last(last_) {}
	
	//construction
	template<typename V>
	vec_op<N+1, vec_op<N,T,U>, V>
	operator<<(const V &v) const {
		return vec_op<N+1, vec_op<N,T,U>, V >(*this, v);
	}
	//...or comma.  same thing.  interchangeable even.
	//the downside is commas won't work in templates (within which my AUTOASSIGN is located...)
	//so you have to replace them with a post-expanded macro like "COMMA" anyways (which is even MORE cumbersome)
//	template<typename V>
//	vec_op<N+1, vec_op<N,T,U>, V>
//	operator,(const V &v) const {
//		return vec_op<N+1, vec_op<N,T,U>, V >(*this, v);
//	}
	
	//element acces
#if 1
	template<int I> typename vec_nth_elem_type<I,N,T,U>::type &elem() {
		return vec_elem<I,N,T,U>(*this);
	}
#else		//lets see of C++ lets us do partial specialization by splitting template args between those of the class and the method...
	template<int I> typename vec_nth_elem_type<I,N,T,U>::type &elem();
	typename vec_nth_elem_type<0,N,T,U>::type &thiselem();
#endif
};
#if 0
template<int I, int N, typename T, typename U>
typename vec_nth_elem_type<I,N,T,U>::type &
vec_op<N,T,U>::elem() {
	return v.next.elem<I-1>;
	//return vec_op<N-1, typename T::nextType, typename T::lastType>::elem<I-1>(v.next)
}
template<int N, int T, int U>
typename vec_this_elem_type<N,T,U>::type &
vec_op<N,T,U>::elem<0>() {
	return v.thiselem();
}

template<int N, typename T, typename U>
inline typename vec_this_elem_type<N,T,U>::type &vec_op<N,T,U>::thiselem() {
	return v.next.thiselem();
}

template<typename U>
inline U &vec_op<1,vec_term,U>::thiselem() {
	return v.last;
}
#endif

typedef vec_op<0, void, void> vec_term;

template<>
class vec_op<0, void, void> {
public:
	enum { dim = 0 };
	typedef vec_op<0, void, void> stype;
	typedef vec_op<0, void, void> dtype;
	
	//construction
	template<typename V>
	vec_op<1, vec_op<0, void, void>, V>
	operator<<(const V &v) const {
		return vec_op<1, vec_term, V >(*this, v);
	}
	//...or comma.  same thing.  interchangeable even.
//	template<typename V>
//	vec_op<1, vec_op<0, void, void>, V>
//	operator,(const V &v) const {
//		return vec_op<1, vec_term, V >(*this, v);
//	}
};

//constructors

//buildingn vectors
//via AUTOASSIGN(var, vec() << xexpr << yexpr << zexpr);

inline vec_term vec() { return vec_term(); }

//shortcut for fixed sizes

template<typename A>
inline vec_op<1, vec_term, A>
vec1(const A &a) {
	return vec() << a;
}

template<typename A,typename B>
inline vec_op<2, vec_op<1, vec_term, A>, B>
vec2(const A &a, const B &b) {
	return vec() << a << b;
}

template<typename A,typename B, typename C>
inline vec_op<3, vec_op<2, vec_op<1, vec_term, A>, B>, C>
vec3(const A &a, const B &b, const C &c) {
	return vec() << a << b << c;
}

template<typename A, typename B, typename C, typename D>
inline vec_op<4, vec_op<3, vec_op<2, vec_op<1, vec_term, A>, B>, C>, D>
vec4(const A &a, const B &b, const C &c, const D &d) {
	return vec() << a << b << c << d;
}

//element access (make this a member?)

template<int N, typename T, typename U>
class vec_this_elem_type {
public:
	typedef typename vec_this_elem_type<N-1, typename T::nextType, typename T::lastType>::type type;
};

template<typename T, typename U>
class vec_this_elem_type<1,T,U> {
public:
	typedef U type;
};

//get the type of the i'th entry of a vector
template<int I, int N, typename T, typename U>
class vec_nth_elem_type {
public:
	typedef typename vec_nth_elem_type<I-1,N-1, typename T::nextType, typename T::lastType>::type type;
};

template<int N, typename T, typename U>
class vec_nth_elem_type<0,N,T,U> {
public:
	typedef typename vec_this_elem_type<N, T, U>::type type;
};

template<int N, typename T, typename U>
typename vec_this_elem_type<N,T,U>::type &
vec_this_elem(vec_op<N,T,U> &v) {
	//symvec.h:185: error: function template partial specialization 'vec_this_elem<1, T, U>' is not allowed
	if (N==1) {
		return v.last;
	} else {
		return vec_this_elem(v.next);
	}
}

//symvec.h:185: error: function template partial specialization 'vec_this_elem<1, T, U>' is not allowed
#if 0
template<typename U>
U &vec_this_elem<1,vec_term,U>(vec_op<1,vec_term,U> &v) {
	return v.last;
}
#endif

template<int I, int N, typename T, typename U>
inline typename vec_nth_elem_type<I,N,T,U>::type &
vec_elem(vec_op<N,T,U> &v) {
	//symvec.h:169: error: function template partial specialization 'vec_elem<0, N, T, U>' is not allowed
	//so until then...
	if (I == 0) {
		return vec_this_elem<N,T,U>(v.last);
	} else {
		return vec_elem<I-1, N-1, typename T::nextType, typename T::lastType>(v.next);
	}
}

//symvec.h:169: error: function template partial specialization 'vec_elem<0, N, T, U>' is not allowed
#if 0
template<int N, typename T, typename U>
inline U &vec_elem<0,N,T,U>(const vec_op<N,T,U> &v) {
	return v.last;
}
#endif

//element acces
//symvec.h:89: error: declaration of 'typename symmath::vec_nth_elem_type<I, N, T, U>::type& symmath::vec_op<N, T, U>::elem()'
//symvec.h:72: error: conflicts with previous declaration 'T symmath::vec_op<N, T, U>::elem'
//so I'm moving it back into the class for now...
#if 0
template<int I, int N, typename T, typename U>
typename vec_nth_elem_type<I,N,T,U>::type &vec_op<N,T,U>::elem() {
	return vec_elem<I,N,T,U>(v);
}
#endif

//ostream

template<int N> inline const char *vec_sep() { return ", "; }
template<> inline const char *vec_sep<1>() { return ""; }

template<int N, typename T, typename U>
inline std::ostream &
operator<<(std::ostream &o, const vec_op<N, T, U> &v) {
	return o << v.next << vec_sep<N>() << v.last;
}

template<>
inline std::ostream &
operator<<(std::ostream &o, const vec_term &v) {
	return o;
}

//debugstr

/*
TODO - get the string to wrap {} only around the OUTERMOST vector
(none of its sub-vectors)
*/
template<int N, typename T, typename U>
inline std::string debugstr(const vec_op<N,T,U> &v) {
	std::ostringstream o;
	o << debugstr(v.next) << vec_sep<N>() << debugstr(v.last);
	return o.str();
}

template<> inline std::string debugstr(const vec_term &v) {
	return "";
}

//diff(vec,var)

template <int N, typename T, typename U>
inline typename vec_op<N, T, U>::dtype
diff(const vec_op<N,T,U> &v, const var &x) {
	return diff(v.next, x) << diff(v.last, x);
}

inline vec_term diff(const vec_term &v, const var &x) {
	return vec_term();
}

//diff(var,vec)

//I need a way of building a vector N T::dtype's for a vector of N vars
// ... but typedefs aren't templated ...
// ... but classes are, and they can have types
template<int N, typename T>
class vec_var_type {
public:
	typedef vec_op<N, typename vec_var_type<N-1,T>::gradtype, typename T::dtype> gradtype;
	typedef vec_op<N, typename vec_var_type<N-1,T>::type, T> type;
};

template<typename T>
class vec_var_type<0,T> {
public:
	typedef vec_term gradtype;
	typedef vec_term type;
};

template <int N, typename U, typename T>
inline typename vec_var_type<N,T>::gradtype
diff(const T &expr, const vec_op<N, U, var> &v) {
	return diff(expr, v.next) << diff(expr, v.last);
}

template<typename T> inline vec_term diff(const T &expr, const vec_term &x) {
	return vec_term();
}

//TODO - diff(vec, vec)

};

#endif	//SYMVEC_H
