/*
	Copyright (c) 2010 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/

#include <iostream>
#include "symmath.h"

//#define USE_MATH_LIB

#ifdef USE_MATH_LIB
#include "symvec.h"
#endif

using namespace symmath;
using namespace std;

//keep this in symmath so function calls will lookup there before global
int main(int argc, char **argv) {
	double _x = 3;
	double _y = 2;
	double _z = 5;
	VAR(x,_x);	//macro aside this would be var x(_x, "x");
	VAR(y,_y);
	VAR(z,_z);
	
	cout << endl << "variables" << endl;
	
#define TESTEXPR(expr)	\
	cout << #expr " = " << (expr) << " = " << debugstr(expr) << endl
#define TESTDIFF(expr,var)	{ TESTEXPR(expr); TESTEXPR(diff(expr,var)); }
#define TESTFDX(func)		TESTDIFF(func(x),x)
	TESTDIFF(x,x);
	TESTDIFF(y,x);

	cout << endl << "expressions" << endl;

	TESTDIFF(x+y,x);
	TESTDIFF(x*y,x);
	TESTEXPR(diff(x*y,y));
	TESTDIFF(4.*x,x);

	TESTEXPR(diff(x/y,y));
	TESTEXPR(diff(1./y,y));
	TESTEXPR(diff(1./(x+y),y));
	TESTEXPR(diff(x/(x+y),y));
	TESTEXPR(diff(y/(x+y),y));
	TESTEXPR(diff(y/y,y));
	TESTEXPR(diff(y*(1./y),y));
	TESTEXPR(diff((x+y)/(x+y),y));
	TESTEXPR(diff((x*y)/(x*x+y*y),y));
	TESTEXPR(diff(x*x*x*x,x));
	TESTEXPR(diff(x*y,x) - diff(x*y,y));
	
	cout << endl << "autoassign" << endl;

#define AUTOASSIGN(x,y)	typeof(y) x = y
	{
		AUTOASSIGN(expr, (x+y)*x);
		TESTDIFF(expr,x);
		TESTEXPR(diff(expr,y));
	}


	cout << endl << "functions" << endl;
	TESTEXPR(cos(cons(2.)));

	TESTFDX(cos);
	TESTFDX(sin);
	TESTDIFF(cos(cos(x)),x);
	TESTDIFF(sin(sin(x)),x);
	TESTFDX(exp);
	TESTFDX(log);
	TESTDIFF(pow(x,2.),x);
	TESTDIFF(pow(2.,x),x);
	TESTFDX(sqrt);
	TESTFDX(cbrt);
	TESTDIFF(pow(x,1./3.),x);
	TESTFDX(cosh);
	TESTFDX(sinh);
	TESTFDX(tanh);
	TESTFDX(asinh);
	TESTFDX(acosh);
	TESTFDX(atanh);

#ifdef USE_MATH_LIB	
	cout << endl << "fixed-size vectors" << endl;
	{
		AUTOASSIGN(r, vec() << x << y << z);
		
		TESTEXPR(r);
		TESTEXPR(diff(vec(),x));
		TESTEXPR(diff(r,x));

		AUTOASSIGN(v, vec() << x*y*z << z/x << x/y);
		TESTEXPR(v);
		TESTEXPR(diff(v,x));
		
		//i think the type tree structure will match regardless of the variables ...
		AUTOASSIGN(expr, x*y*z + y*z + z);
		TESTEXPR(expr);
		TESTEXPR(diff(expr, x));
		TESTEXPR(diff(expr, y));
		TESTEXPR(diff(expr, vec() << x));
		TESTEXPR(diff(expr, vec() << x << y));
		TESTEXPR(diff(expr, r));
		
		AUTOASSIGN(v3, vec3(x*y, z+x, 2.*x - y));
		cout << "elem0 " << debugstr(v3.next.next.last) << endl;
		cout << "elem1 " << debugstr(v3.next.last) << endl;
		cout << "elem2 " << debugstr(v3.last) << endl;

		//not working yet...
//		cout << "elem0 " << debugstr(v3.elem<0>()) << endl;
//		cout << "elem1 " << debugstr(v3.elem<1>()) << endl;
//		cout << "elem2 " << debugstr(v3.elem<2>()) << endl;

		//... and lets not forget -- TODO -- eventually allow dynamic runtime access of elements
	}
	
	//the only way I can think of implementing dynamic-sized vectors with varying types per element
	//is with virtual tables
#endif
	
	return 0;
}
